//
// Copyright (C) 1999 Grace Software
//
// The contents of this file are subject to the Grace Software Public
// License Version 1.0 (the "GracePL"); you may not use this file
// except in compliance with the GracePL. You may obtain a copy of the
// GracePL at http://www.homestead.com/javalog/files/license.html
//
// Software distributed under the GracePL is distributed on an "AS IS"
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the GracePL for the specific language governing rights and
// limitations under the GracePL.
//
// The Original Code is Grace Software's JavaLog code, released March
// 21, 1999.
//
// The Initial Developer of the Original Code is Grace Software.
// Portions created by Grace Software are Copyright (C) 1999 Grace
// Software. All Rights Reserved.
//
// Contributor(s):
//
//
// Copyright (C) 1999 Grace Software
//
// The contents of this file are subject to the Grace Software Public
// License Version 1.0 (the "GracePL"); you may not use this file
// except in compliance with the GracePL. You may obtain a copy of the
// GracePL at http://www.homestead.com/javalog/files/license.html
//
// Software distributed under the GracePL is distributed on an "AS IS"
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the GracePL for the specific language governing rights and
// limitations under the GracePL.
//
// The Original Code is Grace Software's JavaLog code, released March
// 21, 1999.
//
// The Initial Developer of the Original Code is Grace Software.
// Portions created by Grace Software are Copyright (C) 1999 Grace
// Software. All Rights Reserved.
//
// Contributor(s):
//
package grace.io;

import java.io.OutputStream;
import java.io.Writer;

/**
 * This class provides a PrintWriter that can be configured generally
 * by a format String.
 *
 * <P>The formatting operation is as follows.  A format string is
 * dictates the way objects are formatted.  It is composed of the
 * subformat sections.  Each format subsection is separated by a bar
 * ('|') character.  So the general format is:
 *
 * <pre>format: header|begin-block|field|end-block
 *
 * <p>The <i>header</i> format is used only once, at the beginning of
 * the object dump.  It can be used to print one time only header
 * information.
 * 
 * <p>The <i>begin-block</i> format is used just before a sub object
 * is printed.  After using this format, the indent level is increased
 * one level.
 *
 * <p>The <i>field</i> format is used for each field in the object.
 * Formatting is recursive, so if a field is an object, a new sub
 * block is started and the formatting continues recursively.  If the
 * field is a primitive value, it is printed simply.
 *
 * <p>The <i>end-block</i> format is used after a complete sub object has
 * been formatted.
 *
 * <p>The syntax of the each subformat is as follows:
 *
 * <ul>
 * <li>header format:
 *    <ul>
 *    <li>free-text
 *    </ul>
 * <li>begin-block format:
 *    <ul>
 *    <li>free-text
 *    </ul>
 * <li>field format:
 *    <ul>
 *    <li>free-text
 *    <li>%i - indent
 *    <li>%c - field's abreviated class or type
 *    <li>%C - field's full class or type
 *    <li>%n - field's name
 *    <li>%v - field's value
 *    <li>%r - reference id of value
 *    <li>%z - skip the next character if the value printed is the last in list
 *    </ul>
 * <li>end-block
 *    <ul>
 *    <li>free-text
 *    </ul>
 * </ul>
 **/
public class FormattedPrintWriter extends grace.io.PrintWriter {
    private String headerFormat = "";
    private String beginBlockFormat = "{";
    private String fieldFormat = "%c %n=%v;";
    private String endBlockFormat = " }";
    private boolean firstCall = true;

    
    public FormattedPrintWriter(OutputStream stream) {
	super(stream);
    }
    public FormattedPrintWriter(OutputStream stream, boolean autoFlush) {
	super(stream, autoFlush);
    }
    
    public FormattedPrintWriter(Writer writer) {
	super(writer);
    }
    public FormattedPrintWriter(Writer writer, boolean autoFlush) {
	super(writer, autoFlush);
    }


    /**
     * Uses the bar character to separate the beginBlock, field, and
     * endBlock formats.
     **/
    public void setFormat(String format) {
	setFieldFormat("invalid");
	setBeginBlockFormat("invalid");
	setHeaderFormat("invalid");

	// Header
	int bar = format.indexOf('|');
	if (bar != -1) {
	    setHeaderFormat(format.substring(0, bar));

	    // Begin block
	    int start = bar + 1;
	    bar = format.indexOf('|', start);
	    if (bar != -1) {
		setBeginBlockFormat(format.substring(start, bar));

		// Field
		start = bar + 1;
		bar = format.indexOf('|', start);
		if (bar != -1) {
		    setFieldFormat(format.substring(start, bar));
	
		    // End block
		    if (bar + 1 < format.length()) {
			setEndBlockFormat(format.substring(bar + 1));
		    }
		}
	    }
	}
    }


    /**
     * Sets the 4 parts of the format.  See the description of this
     * class above for more information.
     **/
    public void setFormat(
	String header,
	String beginBlock,
	String field,
	String endBlock)
    {
	setHeaderFormat(header);
	setBeginBlockFormat(beginBlock);
	setFieldFormat(field);
	setEndBlockFormat(endBlock);
    }
    
    public void setHeaderFormat(String format) {
	this.headerFormat = format;
    }
    public void setBeginBlockFormat(String format) {
	this.beginBlockFormat = format;
    }
    public void setFieldFormat(String format) {
	this.fieldFormat = format;
    }
    public void setEndBlockFormat(String format) {
	this.endBlockFormat = format;
    }



    /**
     * This formats the object based on the fieldFormat string.
     *
     * <p>This function parses the format:
     * <ul>
     * <li>%i - indent
     * <li>%C - full classname
     * <li>%c - shortened classname
     * <li>%t - shortened JDK classname
     * <li>%n - name
     * <li>%v - value
     * <li>%r - reference number of value
     * <li><i>text</i>
     * </ul>
     **/
    protected void printField(Value value) {
	if (firstCall) {
	    firstCall = false;
	    print(headerFormat);
	}

	boolean inCommand = false;
	boolean skipNextIfLast = false;
	for (int i = 0; i < fieldFormat.length(); ++i) {
	    char c = fieldFormat.charAt(i);

	    if (inCommand) {
		inCommand = false;
		switch (c) {
		    case '%':
			print(c);
			break;
			    
		    case 'i': // append the indent
			indent();
			break;
			    
		    case 'C': // append the class name
			printType(value, false, false);
			break;
			    
		    case 'c': // append the class name
			printType(value, true, true);
			break;
			    
		    case 't': // append the class name
			printType(value, true, false);
			break;
			    
		    case 'n':	// append the name
			printName(value.name);
			break;

		    case 'v':	// append the value
			if (value == null) printNullObject();
			else printValue(value);
			break;
			
		    case 'r':	// append the value
			print(referenceNameOf(value));
			break;
			
		    case 'z':	// append the value
			skipNextIfLast = true;
			break;
			
		    default:
			break;
		}
	    }
	    else if (c == '%') {
		inCommand = true;
	    }
	    else {
		if (skipNextIfLast && value.isLast) {
		    skipNextIfLast = false;
		}
		else print(c);
	    }
	}
    }


    /**
     * This formats the blocking of the object - that is the beginning
     * and ending of the text that surrounds an object.
     *
     * <p>This function parses the format:
     * <ul>
     * <li>%i - indent
     * <li>%c - classname
     * <li>%n - name
     * <li>%v - value
     * <li><i>text</i>
     * </ul>
     **/
    protected void printBlocking(String format) {
	boolean inCommand = false;
	for (int i = 0; i < format.length(); ++i) {
	    char c = format.charAt(i);

	    if (inCommand) {
		inCommand = false;
		switch (c) {
		    case '%':
			print(c);
			break;
			    
		    case 'i': // append the indent
			indent();
			break;
			    
		    default:
			break;
		}
	    }
	    else if (c == '%') {
		inCommand = true;
	    }
	    else {
		print(c);
	    }
	}
    }


    public void printBeginBlock() {
	printBlocking(beginBlockFormat);
    }

    public void printEndBlock() {
	printBlocking(endBlockFormat);
    }


    protected void indent() {
	for (int i = 0; i < indent; ++i) print(indentString);
    }
	    
    protected void newlineAndIndent() {
	print("\n");
	indent();
    }
    
    public void beginList() {
	printBeginBlock();
	++indent;
    }
    
    public void endList() {
	if (--indent < 0) indent = 0;
	printEndBlock();
    }

    public void beginObject() {
	beginList();
    }

    public void endObject() {
	endList();
    }

    public void separateListItems() {
    }
    public void separateObjectItems() {
    }

    protected void printNullString() {
	print("null");
    }
    protected void printNullObject() {
	print("null");
    }

    private int indent = 0;
    private static final String indentString = "    ";
}
