//
// Copyright (C) 1999 Grace Software
//
// The contents of this file are subject to the Grace Software Public
// License Version 1.0 (the "GracePL"); you may not use this file
// except in compliance with the GracePL. You may obtain a copy of the
// GracePL at http://www.homestead.com/javalog/files/license.html
//
// Software distributed under the GracePL is distributed on an "AS IS"
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the GracePL for the specific language governing rights and
// limitations under the GracePL.
//
// The Original Code is Grace Software's JavaLog code, released March
// 21, 1999.
//
// The Initial Developer of the Original Code is Grace Software.
// Portions created by Grace Software are Copyright (C) 1999 Grace
// Software. All Rights Reserved.
//
// Contributor(s):
//
//
// Copyright (C) 1999 Grace Software
//
// The contents of this file are subject to the Grace Software Public
// License Version 1.0 (the "GracePL"); you may not use this file
// except in compliance with the GracePL. You may obtain a copy of the
// GracePL at http://www.homestead.com/javalog/files/license.html
//
// Software distributed under the GracePL is distributed on an "AS IS"
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the GracePL for the specific language governing rights and
// limitations under the GracePL.
//
// The Original Code is Grace Software's JavaLog code, released March
// 21, 1999.
//
// The Initial Developer of the Original Code is Grace Software.
// Portions created by Grace Software are Copyright (C) 1999 Grace
// Software. All Rights Reserved.
//
// Contributor(s):
//
package grace.io;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.InvocationTargetException;
import java.io.OutputStream;
import java.io.Writer;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.sql.SQLException;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Date;

import grace.util.NameMap;

/**
 * This class is similar to the java.io.PrintWriter class except it
 * intelligently prints objects as well as simple types.  It also
 * abstracts some of the formatting features such that derived
 * versions of this class can print various styles of formatting the
 * objects.
 *
 * <p>For example, the LispPrintWriter class will print basic types
 * and objects with a lisp like appearance.  This is useful for the
 * CommandLineSessionManager because it needs to present
 * AreaPoolServer data in a lisp like format.
 *
 * <p>PrintHandlers may be added to the system that print additional
 * types that are not handled correctly by the default printing
 * scheme.  Simply derive from PrintHandler and add an instance of it
 * using the addPrintHandler function.
 *
 * <p>Alternatively, you can create the derived PrintHandler class in
 * the directory grace.io.handlers....  Then, when an object of the
 * type named by the handler is printed, this class will be
 * automatically loaded, instantiated, and used to print the value.
 * For example, the JDK 1.2 java.util.Collection class uses this
 * scheme.  So, a JDK 1.2 specific class called
 * grace.io.handlers.java.util.Collection class exists that is used
 * only when a Collection instance is printed.  This keeps the
 * PrintWriter clases independent of JDK 1.2.
 *
 * <p>To do:
 *
 * <ul>
 *
 * <li>Problem: can't use reference numbers in start of object because
 * then it makes the object referenced.
 *
 * </ul>
 **/
public abstract class PrintWriter extends java.io.PrintWriter {

    public PrintWriter(OutputStream stream) {
	super(stream, false);
	this.autoFlush = false;
    }
    public PrintWriter(OutputStream stream, boolean autoFlush) {
	super(stream, autoFlush);
	this.autoFlush = autoFlush;
    }

    
    public PrintWriter(Writer writer) {
	super(writer, false);
	this.autoFlush = false;
    }
    public PrintWriter(Writer writer, boolean autoFlush) {
	super(writer, autoFlush);
	this.autoFlush = autoFlush;
    }


    /**
     * This is the function a user should call to cause the named
     * object to printed.  Two types of users call this method.  The
     * grace.log.EventFormat class calls this to print a top level
     * object and the code in a printTo method calls this to print a
     * field.  This is why it must keep track of whether this is the
     * first call to it or not.  If it is the first call, it must set
     * the object as the top level object.
     **/
    public void print(String name, Object object) {
	boolean first = firstPrint;
	Value value = new Value(name, object, !first);
	if (first) {
	    value.isFirst = true;
	    value.isLast = true;
	}
	firstPrint = false;
	printField(value);
	if (!first) {
	    autoFlush();
	    firstPrint = false;
	}
    }
    
    /**
     * This is the function a user should call to cause the named object
     * to printed.
     **/
    public void print(String name, int value) {
	print(name, new Integer(value));
    }

    /**
     * This is the function a user should call to cause the named object
     * to printed.
     **/
    public void print(String name, byte value) {
	print(name, new Byte(value));
    }

    /**
     * This is the function a user should call to cause the named object
     * to printed.
     **/
    public void print(String name, char value) {
	print(name, new Character(value));
    }

    /**
     * This is the function a user should call to cause the named object
     * to printed.
     **/
    public void print(String name, float value) {
	print(name, new Float(value));
    }

    /**
     * This is the function a user should call to cause the named object
     * to printed.
     **/
    public void print(String name, double value) {
	print(name, new Double(value));
    }

    /**
     * This is the function a user should call to cause the named
     * object to printed.
     **/
    public void print(String name, boolean value) {
	print(name, new Boolean(value));
    }
    

    /**
     * This meant to be overridden to provide a meaningful way to
     * print the value of a field;
     *
     * @param value of object
     **/
    protected abstract void printField(Value value);
    
    protected void printNullString() {
	printNullObject();
    }
    protected void printNullObject() {
	super.print("null");
    }
    
    protected void printQuoted(String string) {
	if (string == null) printNullString();
	else {
	    beginString();
	    super.print(string);
	    endString();
	}
    }


    /**
     * This prints the given object in one of many ways.  In order, it
     * tries to print the object as a primitive type, print it as a
     * type for which a handler has been added to this PrintWriter,
     * print it as a class that has implemented the printTo function,
     * print the result of calling get functions for each field, or
     * print by calling the toString method.
     **/
    protected void printValue(Value value) {
	if (value.object == null) {
	    printNullObject();
	    return;
	}
	
	//
	// The isArray() check must be performed first so that the
	// array is converted to a list and not simply printed as one
	// of the more basic types.
	// 
	if (value.clazz.isArray()) {
	    print((Object[]) value.object);
	}
	else if (value.clazz == String.class) {
	    printQuoted((String) value.object);
	}
	else if (value.isPrimitive(true)) {
	    super.print(value.toString());
	}
	else if (!alreadyPrinted(value.object)) {
	    rememberPrinted(value.object);
	    
	    PrintHandler handler = getPrintHandler(value.clazz);
	    if (handler != null) {
		handler.handle(this, value);
	    }
	    else {
		Vector accessors = fieldAccessors(value.clazz);
		if (accessors != null) {
		    printWithFieldAccessors(value, accessors);
		}
		else printQuoted(value.toString());
	    }
	}
	else { print(referenceNameOf(value.object)); }
    }


    /**
     * This uses the given Vector<FieldAccessor> to print the fields
     * of the given object using the printField function.
     **/
    public void printWithFieldAccessors(Value parent, Vector accessors) {
	beginObject();
	for (int i = 0; i < accessors.size(); ++i) {
	    FieldAccessor accessor = (FieldAccessor) accessors.elementAt(i);
	    Value child = new Value(
		accessor.name,
		accessor.get(parent.object),
		true);

	    if (i == 0) child.isFirst = true;
	    else if (i == accessors.size() - 1) child.isLast = true;
	    
	    printField(child);
	}
	endObject();
    }


    /**
     * Caches the handler for the given class so that it can be
     * printed by calling the "handle" function on the given
     * PrintHandler.
     *
     * @param type of object to print
     *
     * @param handler on which the "handle" function will be called
     * with a value of type 'type'.
     **/
    public static void addPrintHandler(Class type, PrintHandler handler) {
	printHandlers.put(type, handler);
    }

    /**
     * This finds a PrintHandler by first looking in the handlers
     * subdirectory for the type and then by iterating through the
     * list of added PrintHandlers (added using addPrintHandler) to
     * find a PrintHandler that will handle the requested subtype.  If
     * it still can't find one, it makes a final attempt at looking
     * for a printTo pattern in the class.
     **/
    protected static PrintHandler getPrintHandler(Class toFind) {
	PrintHandler handler = null;

	//
	// First, see if we've got a Handler in the handlers directory
	// to handle objects of this class.
	//
	try {
	    Class handlerClass = Class.forName(
		"grace.io.handlers." + toFind.getClass().getName());
	    handler = (PrintHandler) handlerClass.newInstance();
	}
	catch (Exception e) {}

	if (handler == null) {
	    //
	    // See if there is an exact handler for the object type.
	    //
	    handler = (PrintHandler) printHandlers.get(toFind);

	    if (handler == null) {
		//
		// See if there is a handler that will handle a baser type.
		//
		Enumeration enum = printHandlers.keys();
		while (enum.hasMoreElements()) {
		    Class clazz = (Class) enum.nextElement();
		    if (clazz.isAssignableFrom(toFind))
			return (PrintHandler) printHandlers.get(clazz);
		}
	    }
	}

	//
	// See if there is a printTo method
	//
	if (handler == null) handler = getPrintToHandler(toFind);
	
	return handler;
    }



    /**
     * This returns the abreviated name of the given class. That is,
     * the leading path is stripped off.  It also converts the object
     * versions of the primitive type names into the primitive type
     * names.  If the shortenNonJDKTypes flag is set, every class is
     * converted.
     *
     * @param type whose type name to shorten
     * @param convertNonJDKTypes shorten every type's name rather than JDK's.
     **/
    protected static String shortenedTypeName(
	Class type,
	boolean shortenNonJDKTypes)
    {
	String name = type.getName();

	if (type == Integer.class) name = "int";
	else if (type == Float.class) name = "float";
	else if (type == Long.class) name = "long";
	else if (type == Byte.class) name = "byte";
	else if (type == Boolean.class) name = "boolean";
	else if (type == Character.class) name = "char";
	else if (type == Double.class) name = "double";
	else if (name.startsWith("java") || shortenNonJDKTypes) {
	    name = rightmostName(name);
	}
	else name = type.getName();

	return name;
    }

    /**
     * This returns the rightmost name in a dot '.' separated list of
     * names.
     **/
    protected static String rightmostName(String name) {
	int lastDot = name.lastIndexOf('.');
	return lastDot != -1 ? name.substring(lastDot + 1) : name;
    }

    /**
     * Prints the type of the given object shortened to the rightmost
     * name if desired.
     *
     * @param object whose type to print
     * @param shorten shorten type names to the rightmost name
     * @param shortenAll shorten all type names instead of just JDK names
     **/
    protected void printType(
	Value value,
	boolean shorten,
	boolean shortenAll)
    {
	String typeName;

	if (value.object == null) {
	    typeName = "Object";
	}
	else {
	    if (value.clazz.isArray()) {
		typeName = arrayTypeName(value.clazz, shorten, shortenAll);
	    }
	    else if (shorten) {
		typeName = shortenedTypeName(value.clazz, shortenAll);
	    }
	    else typeName = value.clazz.getName();
	}

	print(typeName);
    }

    protected static String arrayTypeName(
	Class type,
	boolean shorten,
	boolean shortenAll)
    {
	type = type.getComponentType();
	String name = (shorten ? shortenedTypeName(type, shortenAll)
	    : type.getName());

	return name + "[]";
    }


    /**
     * This is called when the name of an object is printed.
     * Currently, this merely prints the object as a string.  This can
     * be overrided to print the name of an object in any way.
     *
     * @param name of object to print
     **/
    protected void printName(Object name) {
	print(name);
    }
    

    /**
     * This returns the reflected method named "printTo(PrintWriter)"
     * on the given clazz.
     **/ 
    protected static Method getPrintMethod(Class objectClass) {
	try {
	    return objectClass.getDeclaredMethod(
		"printTo", printMethodArguments);
	}
	catch (Exception e) { return null; }
    }

    /**
     * Handler class that will invoke the printMethod passed in the
     * constructor when the call to handle is made.  The printMethod
     * is assumed to be on the object that needs to be printed and it
     * is assumed to take a grace.io.PrintWriter as the single
     * argument.
     **/
    static class PrintMethodHandler implements PrintHandler {
	Method printMethod;
	public PrintMethodHandler(Method printMethod) {
	    this.printMethod = printMethod;
	}
	public void handle(PrintWriter writer, Value value) {
	    try { printMethod.invoke(value.object, new Object[] { writer }); }
	    catch (IllegalAccessException e) {}
	    catch (java.lang.reflect.InvocationTargetException e) {}
	}
    }

    
    /**
     * This finds and adds a PrintMethodHandler for the given class that
     * invokes the classes printTo method if available it has one.
     *
     * @param clazz on which to add PrintMethodHandler
     * @return PrintHandler that was added
     **/
    protected static PrintHandler getPrintToHandler(Class clazz) {
	PrintHandler handler = null;
	Method printMethod = getPrintMethod(clazz);
	if (printMethod != null) {
	    handler = new PrintMethodHandler(printMethod);
	    addPrintHandler(clazz, handler);
	}
	
	return handler;
    }


    /**
     * Specifies interface for accessing a field of a class.
     * Anticipated implementors would be a class that accesses via the
     * field directly and one that accesses via a get method.
     **/
    private static abstract class FieldAccessor {
	public String name;
	public FieldAccessor(String name) { this.name = name; }
	public abstract Object get(Object owner);
    }

    /**
     * Access the field directly using the field.get(object) call.
     **/
    private static class DirectFieldAccessor extends FieldAccessor {
	Field field;
	public DirectFieldAccessor(String name, Field field) {
	    super(name);
	    this.field = field;
	}
	
	public Object get(Object owner) {
	    try { return field.get(owner); }
	    catch (IllegalAccessException e) { return "<illegal-access>"; }
	}
    }

    /**
     * Access the field using the reflected get method using
     * method.invoke(object, null) call.
     **/
    private static class MethodFieldAccessor extends FieldAccessor {
	Method accessor;
	public MethodFieldAccessor(String name, Method accessor) {
	    super(name);
	    this.accessor = accessor;
	}
	
	public Object get(Object owner) {
	    try {
		return accessor.invoke(owner, null);
	    }
	    catch (IllegalAccessException e) { return "<illegal-access>"; }
	    catch (java.lang.reflect.InvocationTargetException e) {
		return "<illegal-access>";
	    }
	}
    }


    /**
     * Tries to find and remember the accessors for the given class.
     * It first tries to find the accessors in the classFieldAccessors
     * hashtable.  Then it calls the remmeberAccessorsUsingFields.  If
     * this vector of accessors is zero length, then it calls
     * rememberAccessorsUsingGetMethods.
     *
     * @param clazz for which to remember accessors
     * @return Vector<FieldAccessors>
     **/
    protected static Vector fieldAccessors(Class clazz) {
	Vector accessors = (Vector) classFieldAccessors.get(clazz);
	if (accessors == null) accessors = rememberAccessorsUsingFields(clazz);
	if (accessors.size() == 0) {
	    accessors = rememberAccessorsUsingGetMethods(clazz);
	}

	return accessors;
    }
    
    /**
     * This reflects on the get functions of the given class, finds
     * all of the get functions that look like follow the bean get
     * pattern, and stores them as FieldAccessors in the
     * classFieldAccessors.
     *
     * @return Vector<FieldAccessors>
     **/
    protected static Vector rememberAccessorsUsingGetMethods(Class clazz) {
	Vector remembered = new Vector();
	classFieldAccessors.put(clazz, remembered);
	
	Method[] methods = clazz.getMethods();
	for (int i = 0; i < methods.length; ++i) {
	    String name = methods[i].getName();
	    // lower case the name to get the "field" name
	    String fieldName = Character.toLowerCase(name.charAt(3))
		+ name.substring(4);
	    
	    Class[] arguments = methods[i].getParameterTypes();
	    Class returnType = methods[i].getReturnType();

	    if (name.startsWith("get") && arguments.length == 0
		&& methods[i].getDeclaringClass() != Object.class)
	    {
		remembered.addElement(
		    new MethodFieldAccessor(fieldName, methods[i]));
	    }
	}

	return remembered;
    }
    

    /**
     * This function tries to gain access to the fields of the object
     * using reflection and creates a FieldAccessor for each field
     * depending on if it can be accessed publicly or by a get method.
     *
     * @param clazz to reflect and discover fields and get functions
     * @return Vector<FieldAccessors>
     **/
    protected static Vector rememberAccessorsUsingFields(Class clazz) {
	Vector remembered = new Vector();
	classFieldAccessors.put(clazz, remembered);
	
	//
	// First get all of the declared fields of the class
	// hierarchy.  This must be done by iterating over the
	// declared fields of every class because stupid java doesn't
	// have a way to get all the fields of a class and its base
	// classes - only the public fields.
	//
	// This is done in a separate loop so that the order of the
	// fields is maintained.  In other words, the fields will be
	// accessed eventually from base class to most derived class.
	//
	Vector fields = new Vector();
	Class inspectClass = clazz;
	while (inspectClass != Object.class) {
	    //
	    // Add the declared fields to the top of the list
	    // so the fields will be evaluated from base class to
	    // derived class.
	    //
	    fields = add(inspectClass.getDeclaredFields(), fields);
	    inspectClass = inspectClass.getSuperclass();
	}

	//
	// Now get values for each field.  This assumes a 'get'
	// pattern function that takes no arguments.
	//
	// This sets up access to public fields first if they are
	// available.  If not, this sets up access via a get function.
	//
	for (int i = 0; i < fields.size(); ++i) {
	    Field field = (Field) fields.elementAt(i);
	    if ((field.getModifiers() & Modifier.STATIC) == 0) {
		String fieldName = field.getName();

		//
		// Public or use a get function?
		//
		if ((field.getModifiers() & Modifier.PUBLIC) == 1) {
		    remembered.addElement(
			new DirectFieldAccessor(fieldName, field));
		}
		else {
		    //
		    // Try to call the get method for the field.
		    //
		    String methodName = "get"
			+ Character.toUpperCase(fieldName.charAt(0))
			+ fieldName.substring(1);

		    try {
			Method getMethod = clazz.getMethod(methodName, null);
			remembered.addElement(
			    new MethodFieldAccessor(fieldName, getMethod));
		    }
		    catch (java.lang.NoSuchMethodException e) {} // ignore
		}
	    }
	}

	return remembered;
    }


    /**
     * Sets the DateFormat used for printing all dates.
     **/
    public void setDateFormat(java.text.DateFormat format) {
	dateFormat = format;
    }


    /**
     * Just a utility function to add an array of objects to a vector
     * of objects.
     *
     * @param first array to add to second vector
     * @param second vector appended to first
     * @return first + second
     **/ 
    protected static Vector add(Object[] first, Vector second) {
	Vector result = new Vector(first.length + second.size());

	for (int i = 0; i < first.length; ++i) result.addElement(first[i]);
	for (int i = 0; i < second.size(); ++i)
	    result.addElement(second.elementAt(i));

	return result;
    }


    public boolean prettyOutput() { return pretty; }
    public void setPrettyOutput(boolean pretty) { this.pretty = pretty; }

    public boolean getPrintObjectsOnlyOnce() { return printObjectsOnlyOnce; }
    public void setPrintObjectsOnlyOnce(boolean value) {
	printObjectsOnlyOnce = value;
    }

    public abstract void beginList();
    public abstract void endList();

    public abstract void beginObject();
    public abstract void endObject();

    protected void beginString() { super.print("\""); }
    protected void endString() { super.print("\""); }

    
    /**
     * This records the fact the given object has been printed in the
     * PrintStream and should not be printed again, presumably.
     **/
    protected void rememberPrinted(Object object) {
	if (printObjectsOnlyOnce) printedObjects.put(object);
    }


    protected String referenceNameOf(Object object) {
	return printedObjects.nameOf(object);
    }

    
    /**
     * This indicates that the given object has been printed in
     * the PrintStream and should not be printed again, presumably.
     **/
    protected boolean alreadyPrinted(Object object) {
	if (printObjectsOnlyOnce) return printedObjects.contains(object);
	else return false;
    }

    protected void autoFlush() {
	if (autoFlush) flush();
    }


    /**
     * This method is not meant to be called by the general public.
     * It is public only because it needs to be called by various
     * PrintHandlers in the system.  It is a convience function for
     * printing items of a collection that is indexed.
     **/
    protected void printListElementValue(
	Object value,
	int index,
	int numElements)
    {
	Value element = new Value("[" + index + "]", value, true);
	element.isFirst = index == 0;
	element.isLast = index == numElements - 1;

	printField(element);
    }

    
    protected void print(Object[] array) {
	beginList();
	for (int i = 0; i < array.length; ++i) {
	    printListElementValue(array[i], i, array.length);
	}
	endList();
    }
    

    protected void print(Vector vector) {
	beginList();
	boolean appendedOne = false;
	for (int i = 0; i < vector.size(); ++i) {
	    printListElementValue(vector.elementAt(i), i, vector.size());
	}
	endList();
    }


    protected void print(Hashtable table) {
	Enumeration keys = table.keys();
	Enumeration elements = table.elements();

	beginList();
	boolean first = true;
	while (keys.hasMoreElements() && elements.hasMoreElements()) {
	    Value element = new Value(
		keys.nextElement().toString(),
		elements.nextElement(),
		true);
	    
	    element.isFirst = first;
	    element.isLast = !elements.hasMoreElements();
	    first = false;
	    
	    printField(element);
	}
	endList();
    }



    protected void print(Enumeration elements) {
	beginList();
	int count = 0;
	boolean first = true;
	while (elements.hasMoreElements()) {
	    Value element = new Value(
		"[" + count++ + "]",
		elements.nextElement(),
		true);
	    
	    element.isFirst = first;
	    element.isLast = !elements.hasMoreElements();
	    first = false;
	    
	    printField(element);
	}
	endList();
    }



    /**
     * Just a simple object to differentiate from a ResultSet.
     **/
    private class ResultSetRow {
	public java.sql.ResultSet results;
	public java.sql.ResultSetMetaData meta;
	public int row = 0;
	public ResultSetRow(java.sql.ResultSet resultSet) {
	    results = resultSet;
	    try { meta = results.getMetaData(); }
	    catch (Exception e) { meta = null; }
	}
    }

    public void print(ResultSetRow row) {
	beginList();
	try {
	    if (row.meta != null) {
		for (int i = 0; i < row.meta.getColumnCount(); ++i) {
		    Value value = new Value(
			row.meta.getColumnName(i + 1),
			row.results.getObject(i + 1),
			true);
		    
		    value.isFirst = i == 0;
		    value.isLast = i == row.meta.getColumnCount() - 1;

		    printField(value);
		}
	    }
	}
	catch (SQLException e) { e.printStackTrace(); } // ignore
	endList();
    }
    
    protected void print(java.sql.ResultSet resultSet) {
	beginList();
	int rowNum = 0;
	boolean first = true;
	try {
	    boolean more = resultSet.next();
	    while (more) {
		ResultSetRow row = new ResultSetRow(resultSet);
		Value value = new Value("[" + rowNum + "]", row, true);
		value.isFirst = first;
		more = resultSet.next();
		value.isLast = more;
		printField(value);
		
		++rowNum;
	    }
	}
	catch (SQLException e) { e.printStackTrace(); } // ignore
	endList();
    }


    /**
     * This prints an object that is known only as a remote interface.
     * This means that there are no local fields so it will be printed
     * by only examining the get functions.
     **/ 

    private boolean autoFlush = false;
    private boolean pretty = true;
    private NameMap printedObjects = new NameMap("{ref-", "}");
    private boolean printObjectsOnlyOnce = true;
    private boolean firstPrint = true;
    

    protected DateFormat dateFormat; {
	dateFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
	dateFormat.setTimeZone(java.util.TimeZone.getTimeZone("GMT"));
    }

    
    /**
     * Captures PrintHandlers for the keyed class.
     *
     * template Hashtable<Class, PrintHandler>
     **/
    private static Hashtable printHandlers = new Hashtable();

    /**
     * This is a hashtable of vectors keyed by a Class.  The vector is
     * a list of FieldAccessors for the keyed class.  
     *
     * template Hashtable<Class, Vector<FieldAccessor>>
     **/
    private static Hashtable classFieldAccessors = new Hashtable();

    private static final Class[] printMethodArguments = { PrintWriter.class };
    
    static {
	//
	// Date handler
	//
	addPrintHandler(
	    java.util.Date.class,
	    new PrintHandler() {
	        public void handle(PrintWriter writer, Value value) {
		    writer.printQuoted(
			writer.dateFormat.format((Date) value.object));
		}
	    });

	//
	// Vector handler
	//
	addPrintHandler(
	    Vector.class,
	    new PrintHandler() {
	        public void handle(PrintWriter writer, Value value) {
		    writer.print((Vector) value.object);
		}
	    });

	//
	// Hashtable handler
	//
	addPrintHandler(
	    Hashtable.class,
	    new PrintHandler() {
	        public void handle(PrintWriter writer, Value value) {
		    writer.print((Hashtable) value.object);
		}
	    });

	//
	// Enumeration handler
	//
	addPrintHandler(
	    Enumeration.class,
	    new PrintHandler() {
	        public void handle(PrintWriter writer, Value value) {
		    writer.print((Enumeration) value.object);
		}
	    });

	//
	// ResultSet handler
	//
	addPrintHandler(
	    java.sql.ResultSet.class,
	    new PrintHandler() {
	        public void handle(PrintWriter writer, Value value) {
		    writer.print((java.sql.ResultSet) value.object);
		}
	    });

	//
	// ResultSetRow handler
	//
	addPrintHandler(
	    ResultSetRow.class,
	    new PrintHandler() {
	        public void handle(PrintWriter writer, Value value) {
		    writer.print((ResultSetRow) value.object);
		}
	    });
    }
}
