//
// Copyright (C) 1999 Grace Software
//
// The contents of this file are subject to the Grace Software Public
// License Version 1.0 (the "GracePL"); you may not use this file
// except in compliance with the GracePL. You may obtain a copy of the
// GracePL at http://www.homestead.com/javalog/files/license.html
//
// Software distributed under the GracePL is distributed on an "AS IS"
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the GracePL for the specific language governing rights and
// limitations under the GracePL.
//
// The Original Code is Grace Software's JavaLog code, released March
// 21, 1999.
//
// The Initial Developer of the Original Code is Grace Software.
// Portions created by Grace Software are Copyright (C) 1999 Grace
// Software. All Rights Reserved.
//
// Contributor(s):
//
//
// Copyright (C) 1999 Grace Software
//
// The contents of this file are subject to the Grace Software Public
// License Version 1.0 (the "GracePL"); you may not use this file
// except in compliance with the GracePL. You may obtain a copy of the
// GracePL at http://www.homestead.com/javalog/files/license.html
//
// Software distributed under the GracePL is distributed on an "AS IS"
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the GracePL for the specific language governing rights and
// limitations under the GracePL.
//
// The Original Code is Grace Software's JavaLog code, released March
// 21, 1999.
//
// The Initial Developer of the Original Code is Grace Software.
// Portions created by Grace Software are Copyright (C) 1999 Grace
// Software. All Rights Reserved.
//
// Contributor(s):
//
package grace.log;

import java.io.StringWriter;
import java.io.PrintWriter;
import java.text.DateFormat;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.TimeZone;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Hashtable;

import grace.util.Properties;
import grace.util.Tokenizer;


/**
 * This class simply captures all of the formatting functionality into
 * a class rather than have it in a specific Handler.  This is to
 * allow any Handler classes to format all or portions of Event text
 * even though the PrintHandler class is the only class producing
 * purely textual output.
 *
 * <p>Typically, a Handler, such as PrintHandler, will instantiate an
 * EventFormat for each different format it needs to support.  In the
 * case of PrintHandler, this is only one.  Then, the Handler calls
 * the format(Event) function each time it wants to format an event.
 *
 * <pre>
 *	public void log(Event event) {
 *	    EventFormat format = new EventFormat();
 *	    System.out.println(format.format(event));
 *      }
 * </pre>
 *
 * <h3>Formatting</h3>
 * <p><i>event-format</i> is composed of a concatenation
 * of the following items:
 * 
 *	<ul>
 *	<li>%t is time
 *	<li>%(<i>time-format</i>)t is time with custom format
 *	<li>%r is relative time 
 *	<li>%e is event type
 *	<li>%h is the thread name(s)
 *	<li>%(<i>thread-format</i>)h is the thread name(s) with custom format
 *	<li>%c is classname
 *	<li>%n is event sequence number
 *	<li>%(<i>class-format</i>)c is classname with custom format
 *	<li>%f is function name
 *	<li>%F is file name
 *	<li>%(<i>filename-format</i>)F is file name with custom format
 *	<li>%l is line number in function
 *	<li>%m is message
 *	<li>%(<i>message-format</i>)m is message with custom format
 *	<li>%o is object
 *	<li>%(<i>object-format</i>)o is object with custom format
 *	<li><i>text</i>
 *	</ul>
 *
 * <p><i>time-format</i> is one of the following:
 *	<ul>
 *	<li>short
 *	<li>medium
 *	<li>long
 *	<li>full
 *	<li>24
 *	<li><i>java.text.SimpleDateFormat</i>
 *	</ul>
 *
 * <p><i>relative-time-format</i> is one of the following:
 *	<ul>
 *	<li>clock
 *	<li>days
 *	<li>hours
 *	<li>minutes
 *	<li>seconds
 *	</ul>
 *
 * <p><i>thread-format</i> is:
 *	<ul>
 *	<li>%g is the thread's group name
 *	<li>%G are all of the thread's group names delimited by %punct
 *	<li>%n is the name of the thread
 *	<li>%punct sets the delimiter of %G
 *	<li><i>plain-text</i>
 *	</ul>
 *
 * <br>Note: to specify the entire hierarchy of group names, use the
 * syntax: <code>%g.%g</code> where the '.', which can be any
 * character, will be used to delimit each thread group name.
 *	
 *
 * <p><i>class-format</i> is:
 *	<ul>
 *	<li>short - class name without the package name
 *	<li>full - class name with the package name
 *	</ul>
 *
 * <p><i>filename-format</i> is:
 *	<ul>
 *	<li>short - filename without the path
 *	<li>full - filename with the path
 *	</ul>
 *
 * <p><i>message-format</i> is:
 *	<ul>
 *	<li>%m is the message
 *	<li><i>plain-text</i>
 *	</ul>
 *    
 * <p><i>object-format</i> is:
 *	<ul>
 *	<li>java with grace.io.JavaPrintWriter
 *	<li>Java with grace.io.JavaPrintWriter pretty
 *	<li>lisp with grace.io.LispPrintWriter
 *	<li>Lisp with grace.io.LispPrintWriter
 *	<li>object-format-string using grace.io.FormattedPrintWriter
 *	</ul>
 *
 * <p><i>exception-format</i> is:
 *	<ul>
 *	<li>%c exception classname
 *	<li>%m message
 *	<li>%s stack trace
 *	<li><i>plain-text</i>
 *	</ul>
 *
 * In this scheme, the current format would be:
 *	<code>"%t(%r):%n: %e: %c.%f:%l: %m: %o\n"</code>
 *
 **/
public class EventFormat {
    public static final String rcsid = "$Id: EventFormat.java,v 1.14 1999/10/11 00:09:59 cbladon Exp $";

    private static Properties properties = new Properties(); static {
	properties.loadSystem();
	properties.integrateWithSystemProperties();
    }

    // Relative time format
    private static String defaultRelativeTimeFormat = properties.get(
	"log.time.relative", "clock");

    // Time format
    private static DateFormat defaultTimeFormat; static {
	defaultTimeFormat = getDateFormat(
	    properties.get("log.time.format", "24"));
	defaultTimeFormat.setTimeZone(
	    java.util.TimeZone.getTimeZone(
		properties.get("log.time.zone", "GMT")));
    }

    // Object time format
    private static DateFormat defaultObjectTimeFormat; static {
	defaultObjectTimeFormat = getDateFormat(
	    properties.get("log.object.time.format", "24"));
	defaultObjectTimeFormat.setTimeZone(
	    TimeZone.getTimeZone(
		properties.get("log.object.time.zone", "GMT")));
    }

    // Exception format
    private static String defaultExceptionFormat = expandSpecialChars(
	properties.get("log.exception.format", ": %s"));

    // Message format
    private static String defaultMessageFormat = expandSpecialChars(
	properties.get("log.message.format", "%m"));

    // Object format
    private static String terseJavaFormat = ":|{| %c %n = %v;| }";
    private static String prettyJavaFormat =":|{\n|%i%c %n = %v;\n|%i}";
    private static String lispFormat = ":|(%v%z,%z |)";
    private static String defaultObjectFormat = expandSpecialChars(
	properties.get("log.object.format", "java"));

    // Thread format
    private static String defaultThreadFormat = expandSpecialChars(
	properties.get("log.thread.format", "%G.%n"));

    // General Format
    private static String defaultFormat = expandSpecialChars(
	properties.get(
	    "log.format",
	    "%t(%r):%n: %e: %h: %c.%f:%l%(: %m)m%o\n\n"));

    private static boolean useColors = properties.get(
	new String[] {"log.color", "log.colors", "log.colour", "log.colours"},
	"false").equals("true");


    public static final String NO_COLOR = "";
    public static final char ESC = '\u001B';
    public static final String PLAIN = ESC + "[00m";
    public static final String BRIGHT = ESC + "[1m";
    public static final String UNDERLINE = ESC + "[4m";
    public static final String HIGHLIGHT = ESC + "[7m";
    public static final String BLACK = ESC + "[30m";
    public static final String RED = ESC + "[31m";
    public static final String GREEN = ESC + "[32m";
    public static final String YELLOW = ESC + "[33m";
    public static final String BLUE = ESC + "[34m";
    public static final String PURPLE = ESC + "[35m";
    public static final String CYAN = ESC + "[36m";
    public static final String WHITE = ESC + "[37m";
    public static final String BLACK_BACKGROUND = ESC + "[40m";
    public static final String RED_BACKGROUND = ESC + "[41m";
    public static final String GREEN_BACKGROUND = ESC + "[42m";
    public static final String YELLOW_BACKGROUND = ESC + "[43m";
    public static final String BLUE_BACKGROUND = ESC + "[44m";
    public static final String PURPLE_BACKGROUND = ESC + "[45m";
    public static final String CYAN_BACKGROUND = ESC + "[46m";
    public static final String WHITE_BACKGROUND = ESC + "[47m";

    private static Hashtable colorTable = new Hashtable(); static {
	colorTable.put("plain", PLAIN);
	colorTable.put("bright", BRIGHT);
	colorTable.put("underline", UNDERLINE);
	colorTable.put("highlight", HIGHLIGHT);
	colorTable.put("black", BLACK);
	colorTable.put("red", RED);
	colorTable.put("green", GREEN);
	colorTable.put("yellow", YELLOW);
	colorTable.put("blue", BLUE);
	colorTable.put("purple", PURPLE);
	colorTable.put("cyan", CYAN);
	colorTable.put("white", WHITE);
	colorTable.put("black-bg", BLACK_BACKGROUND);
	colorTable.put("red-bg", RED_BACKGROUND);
	colorTable.put("green-bg", GREEN_BACKGROUND);
	colorTable.put("yellow-bg", YELLOW_BACKGROUND);
	colorTable.put("blue-bg", BLUE_BACKGROUND);
	colorTable.put("purple-bg", PURPLE_BACKGROUND);
	colorTable.put("cyan-bg", CYAN_BACKGROUND);
	colorTable.put("white-bg", WHITE_BACKGROUND);
    }


    private static final Date startTime = new Date();	// used for relative
    private static final NumberFormat numberFormat =
	NumberFormat.getInstance();
    private static final NumberFormat twoDigitFormat = new DecimalFormat("00");


    
    private String	format;
    private int		minBufferSize = 128;

    private DateFormat	timeFormat = defaultTimeFormat;
    private String	relativeTimeFormat = defaultRelativeTimeFormat;
    private String	exceptionFormat = defaultExceptionFormat;
    private String	messageFormat = defaultMessageFormat;
    private String	objectFormat = defaultObjectFormat;
    private DateFormat	objectTimeFormat = defaultObjectTimeFormat;
    private String	threadFormat = defaultThreadFormat;

    private String	timeColor = NO_COLOR;
    private String 	relativeTimeColor = NO_COLOR;
    private String 	exceptionColor = NO_COLOR;
    private String 	messageColor = NO_COLOR;
    private String 	objectColor = NO_COLOR;
    private String 	threadColor = NO_COLOR;
    private String 	classnameColor = NO_COLOR;
    private String 	functionColor = NO_COLOR;

    private String 	resetColor = NO_COLOR;
    private Hashtable 	eventColors = new Hashtable();
    private Hashtable 	lineColors = new Hashtable();



    /**
     * This constructs an EventFormat with the default formatting.
     * The default formatting set in code, but can be overridden on
     * the command line by setting the appropriate System properties.
     *
     * @see grace.log.Log
     **/
    public EventFormat() {
	this(defaultFormat);
    }
    
    /**
     * This constructs an EventFormat with custom formatting.  This
     * formatting overrides the formatting set in code and the
     * formatting specified on the command line by setting.
     *
     * @see grace.log.Log
     **/
    public EventFormat(String format) {
	this.format = format;
	useColors(useColors);
    }


    /**
     * This is function creates the buffer in which an event will be
     * written by the <code>format(Event)</code> function.  It uses
     * the minBufferSize parameter (set in the bottom of the
     * <code>format(Event)</code> function) to create the StringBuffer
     * with the right capacity.
     *
     * @return StringBuffer sized to anticipate the size of the formatted event
     **/
    private StringBuffer createBuffer() {
	return minBufferSize == 0
	    ? new StringBuffer() : new StringBuffer(minBufferSize);
    }


    /**
     * This converts and returns the given <code>event</code> as a
     * String using a default format or the format set in the
     * EventFormat constructor.
     *
     * <p>This function parses the format:
     * <ul>
     * <li>%t   - time
     * <li>%e   - event type
     * <li>%r|R - relative time
     * <li>%h|H - thread name
     * <li>%n   - event number
     * <li>%c   - originating class name
     * <li>%f   - originating function name
     * <li>%F   - originating file name
     * <li>%l   - originating line number
     * <li>%m   - message
     * <li>%o   - object or exception
     * <li><i>text</i>
     * </ul>
     *
     * @param event to convert into a String using the set format
     *
     * @return string result of converting event using format
     **/
    public String format(Event event) {
	StringBuffer buffer = createBuffer();
	String subFormat = null;

	String lineColor = (String) lineColors.get(event.type);
	if (lineColor != null) buffer.append(lineColor);
		
	boolean inCommand = false;
	for (int i = 0; i < format.length(); ++i) {
	    char c = format.charAt(i);

	    if (inCommand) {
		if (lineColor != null) buffer.append(lineColor);

		inCommand = false;
		switch (c) {
		    case '%':
			buffer.append(c);
			break;
			
		    case 't':	// Time
			formatTime(buffer, subFormat, event.time);
			break;
			
		    case 'e':	// Event type
			String color = (String) eventColors.get(event.type);
			if (color != null) buffer.append(color);
			buffer.append(event.type);
			buffer.append(resetColor);
			break;
			
		    case 'r':	// relative time (short)
		    case 'R':	// Relative Time (long)
			formatRelativeTime(buffer, subFormat, event.time);
			break;
			
		    case 'h':	// Thread name (short)
		    case 'H':	// Thread name (long)
			formatThreadName(buffer,
			    subFormat, event.threadNames);
			break;
			
		    case 'n':	// Number
			buffer.append(event.number);
			break;
			
		    case 'c':	// Classname (short)
		    case 'C':	// Classname (long)
			buffer.append(classnameColor);
			buffer.append(event.position.getShortClassname());
			buffer.append(resetColor);
			break;
			
		    case 'F':	// File name
			buffer.append(event.position.getFilename());
			break;
			
		    case 'f':	// Function name
			buffer.append(functionColor);
			buffer.append(event.position.getFunction());
			buffer.append(resetColor);
			break;
			
		    case 'l':	// Line number
			buffer.append(event.position.getLineNumberString());
			break;
			
		    case 'm':	// Message (short)
		    case 'M':	// Message (long)
			//
			// Only print the message if there was no object
			// specified in this event.
			//
			if (event.message != null && !event.objectWasGiven())
			    formatMessage(buffer, subFormat, event.message);
			break;
			
		    case 'o':	// Object (short)
		    case 'O':	// Object (long)
			if (event.objectWasGiven())
			    formatObject(
				buffer,
				subFormat,
				(event.message == null ? "" : event.message),
				event.object);
			break;
			
		    case '(':	// start sub format
			{
			    //
			    // This should probably be a separate
			    // function, but returning the counter and
			    // adjusting the index i is too much of a
			    // hassle right now.
			    //
			    StringBuffer subBuffer = new StringBuffer();
			    boolean escapeSeen = false;
	
			    while (++i < format.length()) {
				char cc = format.charAt(i);

				if (cc == '\\') escapeSeen = true;
				else if (cc == ')' && !escapeSeen) break;
				else {
				    subBuffer.append(cc);
				    escapeSeen = false;
				}
			    }

			    subFormat = subBuffer.toString();
			}
			inCommand = true;
			break;
			
		    default:
			break;
		}

		if (lineColor != null) buffer.append(lineColor);
	    }
	    else if (c == '%') {
		inCommand = true;
		subFormat = null;
	    }
	    else {
		buffer.append(c);
	    }
	}

	//
	// To optimize the creation of the buffer, this next few lines
	// simply remembers the buffer length so it can use it the
	// next time this format is used.
	//
	if (minBufferSize < buffer.length()) {
	    minBufferSize = buffer.length() + 16;
	}

	buffer.append(resetColor);

	return buffer.toString();
    }


    /**
     * This formats the given exception.
     *
     * <p>This function parses the format:
     * <ul>
     * <li>%c - exception classname
     * <li>%m - exception message
     * <li>%s - exception stack trace
     * <li><i>text</i>
     * </ul>
     *
     * <p>This needs work.
     *
     * <ul>
     *
     * <li>Currently, there is no way to do a short version of the
     * exception.  Also, the long exception is always printing an
     * extra newline at the end of the stack trace.
     *
     * <li>The exception stack trace always includes the class name
     * and the message also.  This should be parsed out.
     *
     * </ul>
     **/
    protected void formatException(
	StringBuffer buffer,
	Exception exception)
    {
	String subFormat = exceptionFormat;

	buffer.append(exceptionColor);
	
	boolean inCommand = false;
	for (int i = 0; i < subFormat.length(); ++i) {
	    char c = subFormat.charAt(i);

	    if (inCommand) {
		inCommand = false;
		switch (c) {
		    case '%':
			buffer.append(c);
			break;
			    
		    case 'c': // append the class name
			buffer.append(exception.getClass().getName());
			break;
			    
		    case 'm':	// append the exception message
			if (exception.getMessage() != null)
			    buffer.append(exception.getMessage());
			break;

		    case 's':	// append the stack trace
			{
			    StringWriter stringWriter = new StringWriter();
			    PrintWriter writer = new PrintWriter(stringWriter);
			    exception.printStackTrace(writer);
			    buffer.append(
				eliminateEndingNewlines(
				    stringWriter.toString()));
			    writer.close();
			}
			break;
			
		    default:
			break;
		}
	    }
	    else if (c == '%') {
		inCommand = true;
	    }
	    else {
		buffer.append(c);
	    }
	}

	buffer.append(resetColor);
    }


    protected String eliminateEndingNewlines(String source) {
	int end = source.length();
	while (end > 0 && source.charAt(end - 1) == '\n') --end;

	return (end > 0) ? source.substring(0, end) : source;
    }
	
    
    /**
     * This formats a string representation of the given message.  The
     * setMessageFormat() function always overrides subFormats.
     *
     * <p>This function parses the format:
     * <ul>
     * <li>%m - message
     * <li><i>text</i>
     * </ul>
     *
     * @param buffer to which to write object
     * @param format from which to format this object
     * @param message object to format
     **/
    protected void formatMessage(
	StringBuffer buffer,
	String subFormat,
	String message)
    {
	if (messageFormat == defaultMessageFormat) {
	    if (subFormat == null) subFormat = messageFormat;
	}
	else subFormat = messageFormat;
	
	buffer.append(messageColor);
	
	boolean inCommand = false;
	for (int i = 0; i < subFormat.length(); ++i) {
	    char c = subFormat.charAt(i);

	    if (inCommand) {
		inCommand = false;
		switch (c) {
		    case '%':
			buffer.append(c);
			break;
			    
		    case 'm':	// append the message itself
			buffer.append(message);
			break;
			    
		    default:
			break;
		}
	    }
	    else if (c == '%') {
		inCommand = true;
	    }
	    else {
		buffer.append(c);
	    }
	}
	
	buffer.append(resetColor);
    }



    /**
     * This function writes the given object to the given buffer using
     * the grace.io.JavaPrintWriter.
     *
     * @param buffer to which this function should append
     * @param object to append to buffer
     * @param name value's name (should never be null)
     * @param pretty flag given to JavaPrintWriter to control length of output
     **/
    protected void formatFormattedObject(
	String subFormat,
	StringBuffer buffer,
	String name,
	Object object)
    {
	StringWriter stringWriter = new StringWriter();
	grace.io.FormattedPrintWriter writer =
	    new grace.io.FormattedPrintWriter(stringWriter, false);
	writer.setDateFormat(objectTimeFormat);
	writer.setFormat(subFormat);
	
	writer.print(name, object);
	writer.close();
	try { stringWriter.close(); }
	catch (Exception e) { Log.internal(e); }

	buffer.append(stringWriter.toString());
    }


    /**
     * This function writes the given object to the given buffer using
     * the given format string.
     *
     * <p>This function parses the format:
     * <ul>
     * <li>%l - using LispPrintWriter
     * <li>%L - using LispPrintWriter pretty
     * <li>%j - using JavaPrintWriter
     * <li>%J - using JavaPrintWriter pretty
     * <li>%o - using default PrintWriter (java)
     * <li><i>text</i>
     * </ul>
     *
     * @param buffer to which this function should append
     * @param format containing text and escape sequences
     * @param name value's name (should never be null)
     * @param object to append to buffer
     **/
    protected void formatGenericObject(
	StringBuffer buffer,
	String subFormat,
	String name,
	Object object)
    {
	if (objectFormat == defaultObjectFormat) {
	    if (subFormat == null) subFormat = objectFormat;
	}
	else subFormat = objectFormat;
	
	buffer.append(objectColor);
	
	if (subFormat.equals("java")) {
	    formatFormattedObject(terseJavaFormat, buffer, name, object);
	}
	else if (subFormat.equals("Java")) {
	    formatFormattedObject(prettyJavaFormat, buffer, name, object);
	}
	else if (subFormat.equals("lisp")) {
	    formatFormattedObject(lispFormat, buffer, name, object);
	}
	else if (subFormat.equals("Lisp")) {
	    formatFormattedObject(lispFormat, buffer, name, object);
	}
	else {
	    formatFormattedObject(subFormat, buffer, name, object);
	}

	buffer.append(resetColor);
    }


    
    /**
     * This formats a string representation of the given object.  It
     * uses one of the grace.io.PrintWriters to convert the object
     * to a string.
     *
     * @param buffer to which to write object
     * @param format from which to format this object
     * @param message to use as value's name (should never be null)
     * @param object object to format
     **/
    protected void formatObject(
	StringBuffer buffer,
	String format,
	String message,
	Object object)
    {
	if (object instanceof Throwable) {
	    formatException(buffer, (Exception) object);
	}
	else {
	    formatGenericObject(buffer, format, message, object);
	}
    }
    


    /**
     * This writes the given Vector of names to the given buffer
     * assuming the format contains thread name escapes.
     *
     * <p>This function parses the format:
     * <ul>
     * <li>%. - where . is any punctuation
     * <li>%n - thread name
     * <li>%g - thread's group name
     * <li>%G - all thread's groups separated by %. punctuation
     * <li><i>text</i>
     * </ul>
     *
     * @param buffer to which the names will be written
     * @param format containing thread name escape sequences
     * @param names to be written to buffer according to format
     **/
    protected void formatThreadName(
	StringBuffer buffer,
	String subFormat,
	Vector names)
    {
	if (threadFormat == defaultThreadFormat) {
	    if (subFormat == null) subFormat = threadFormat;
	}
	else subFormat = threadFormat;
	
	buffer.append(threadColor);
	
	char groupDelimiter = '.';
	boolean inCommand = false;
	for (int i = 0; i < subFormat.length(); ++i) {
	    char c = subFormat.charAt(i);

	    if (inCommand) {
		inCommand = false;
		switch (c) {
		    case '%':
			buffer.append(c);
			break;
			    
		    case 'n':	// append the thread name only
			if (names.size() >= 1) {
			    buffer.append(
				(String) names.elementAt(names.size() - 1));
			}
			break;
			    
		    case 'g':	// append thread's group name
			if (names.size() >= 2) {
			    buffer.append(
				(String) names.elementAt(names.size() - 2));
			}
			break;

		    case 'G':	// append all of the thread's group names
			for (int j = 0; j < names.size() - 1; ++j) {
			    if (j != 0) buffer.append(groupDelimiter);
			    buffer.append((String) names.elementAt(j));
			}
			break;
			
		    default:	// set the delimiter separates the group names
			groupDelimiter = c;
			break;
		}
	    }
	    else if (c == '%') {
		inCommand = true;
	    }
	    else {
		buffer.append(c);
	    }
	}

	buffer.append(resetColor);
    }
    

    /**
     * This writes the difference in time between <code>now</code> and
     * the start of the application to the given <code>buffer</code
     * according to the given <code>format</code>.
     *
     * <p>This function parses the format:
     * <ul>
     * <li>clock
     * <li>days
     * <li>hours
     * <li>minutes
     * <li>seconds
     * </ul>
     *
     * @param buffer to which the relative time will be written
     * @param format containing relative time format instructions
     *
     * @param now from which the application start time is subtracted
     * to form the difference
     **/
    protected void formatRelativeTime(
	StringBuffer buffer,
	String subFormat,
	Date now)
    {
	if (relativeTimeFormat == defaultRelativeTimeFormat) {
	    if (subFormat == null) subFormat = relativeTimeFormat;
	}
	else subFormat = relativeTimeFormat;
	
	buffer.append(relativeTimeColor);
	
	if (subFormat != null && !subFormat.equals("")) {
	    double difference = (double) (now.getTime() - startTime.getTime());
	    if (subFormat.startsWith("clock")) {
		int days = (int) (difference / 86400000);
		int hours = (int) ((difference % 86400000) / 3600000);
		int minutes = (int) ((difference % 3600000) / 60000);
		double seconds = (difference % 60000) / 1000.0;

		if (days > 0) {
		    buffer.append(twoDigitFormat.format(days));
		    buffer.append(":");
		    buffer.append(twoDigitFormat.format(hours));
		    buffer.append(":");
		    buffer.append(twoDigitFormat.format(minutes));
		    buffer.append(":");
		    buffer.append(numberFormat.format(seconds));
		}
		else if (hours > 0) {
		    buffer.append(twoDigitFormat.format(hours));
		    buffer.append(":");
		    buffer.append(twoDigitFormat.format(minutes));
		    buffer.append(":");
		    buffer.append(numberFormat.format(seconds));
		}
		else if (minutes > 0) {
		    buffer.append(twoDigitFormat.format(minutes));
		    buffer.append(":");
		    buffer.append(numberFormat.format(seconds));
		}
		else {
		    buffer.append(numberFormat.format(seconds));
		}
	    }
	    else if (subFormat.startsWith("day")) {
		difference /= 86400000.0;
		buffer.append(numberFormat.format(difference));
	    }
	    else if (subFormat.startsWith("min")) {
		difference /= 60000.0;
		buffer.append(numberFormat.format(difference));
	    }
	    else if (subFormat.startsWith("sec")) {
		difference /= 1000.0;
		buffer.append(numberFormat.format(difference));
	    }
	    else { // default to relative hours
		difference /= 3600000.0;
		buffer.append(numberFormat.format(difference));
	    }
	}
	buffer.append(resetColor);
    }


    /**
     * This returns a string representation of the current time,
     * formatted according to the previously configured system
     * properties.
     *
     * <p>This function parses the format:
     * <ul>
     * <li>short  - using java.text.Format.getTimeDateInstance(SHORT);
     * <li>medium - using java.text.Format.getTimeDateInstance(MEDIUM);
     * <li>long   - using java.text.Format.getTimeDateInstance(LONG);
     * <li>full   - using java.text.Format.getTimeDateInstance(FULL);
     * <li>format - using java.text.SimpleDateFormat
     * </ul>
     *
     **/
    protected void formatTime(
	StringBuffer buffer,
	String subFormat,
	Date now)
    {
	DateFormat toUse = timeFormat;
	
	if (timeFormat == defaultTimeFormat && subFormat != null) {
	    toUse = getDateFormat(subFormat);
	}
	buffer.append(timeColor);
	buffer.append(toUse.format(now));
	buffer.append(resetColor);
    }


    

    /**
     * Utility function used by <code>formatTime(...)</code> to find
     * and return java.text.DateFormat for the given name.
     *
     * @param name { short, medium, long, full, <SimpleDateFormat> }
     *
     * @return DateFormat corresponding to name
     **/
    protected static DateFormat getDateFormat(String name) {
	DateFormat format = null;
	
	if (name.equals("short")) {
	    format = DateFormat.getDateTimeInstance(
		DateFormat.SHORT,
		DateFormat.SHORT);
	}
	else if (name.equals("medium")) {
	    format = DateFormat.getDateTimeInstance(
		DateFormat.MEDIUM,
		DateFormat.MEDIUM);
	}
	else if (name.equals("long")) {
	    format = DateFormat.getDateTimeInstance(
		DateFormat.LONG,
		DateFormat.LONG);
	}
	else if (name.equals("full")) {
	    format = java.text.DateFormat.getDateTimeInstance(
		DateFormat.FULL,
		DateFormat.FULL);
	}
	else if (name.equals("24")) {
	    format = new SimpleDateFormat("MM/dd/yy HH:mm:ss");
	}
	else {
	    // assume it is a SimpleDateFormat pattern
	    format = new SimpleDateFormat(name);
	}

	return format;
    }


    /**
     * This expands newlines and tab escapes into the actual
     * character.  This needed so that the user can enter these
     * special characters on the command line and have them work like
     * in Java.  It handles null string by returning null.
     *
     * @param string containing potential escaped newlines and tabs
     *
     * @return the string with the escaped chars replaced by actual
     * chars or null if string is null.
     **/
    protected static String expandSpecialChars(String string)
    {
	if (string != null) {
	    StringBuffer buffer = new StringBuffer(string.length());

	    boolean inEscape = false;
	    for (int i = 0; i < string.length(); ++i) {
		char c = string.charAt(i);

		if (inEscape) {
		    inEscape = false;
		    switch (c) {
			case 'n': buffer.append('\n'); break;
			case 't': buffer.append('\t'); break;
			default:  buffer.append(c); break;
		    }
		}
		else if (c == '\\') inEscape = true;
		else buffer.append(c);
	    }

	    return buffer.toString();
	}

	return null;
    }

    public void setObjectFormat(String format) { objectFormat = format; }
    public void setObjectTimeFormat(DateFormat format) {
	objectTimeFormat = format;
    }
    public void setTimeFormat(DateFormat format) { timeFormat = format; }
    public void setRelativeTimeFormat(String format) {
	relativeTimeFormat = format;
    }
    public void setExceptionFormat(String format) { exceptionFormat = format; }
    public void setMessageFormat(String format) { messageFormat = format; }
    public void setThreadFormat(String format) { threadFormat = format; }

    public void useColors(boolean colors) {
	if (colors) {
	    timeColor = BRIGHT;
	    resetColor = PLAIN;
	    eventColors.put(Log.ERROR, HIGHLIGHT + RED);
	    eventColors.put(Log.WARNING, HIGHLIGHT);
	    lineColors.put(Log.ERROR, BRIGHT + RED);
	    lineColors.put(Log.WARNING, BRIGHT);
	}
	else {
	    timeColor = NO_COLOR;
	    resetColor = NO_COLOR;
	    eventColors.clear();
	}
    }

    /**
     * Takes a space separated list of string colors and returns a
     * composite color that can be used to set any elements color of
     * this class.  For example, given a property, one can call
     * setEventColor(colorListToColor(property));
     **/
    public String colorListToColor(String colorList) {
	Tokenizer tokens = new Tokenizer(colorList);
	StringBuffer color = new StringBuffer();
	String item;
	while ((item = tokens.getToken()) != null) {
	    item = (String) colorTable.get(item);
	    if (item != null) color.append(item);
	}
	
	return color.toString();
    }
    
    /** Sets the color of the formatted time. **/
    public void setTimeColor(String color) {
	timeColor = color;
	resetColor = PLAIN;
    }
    
    /** Sets the color of the formatted relative time. **/
    public void setRelativeTimeColor(String color) {
	relativeTimeColor = color;
	resetColor = PLAIN;
    }
    
    /** Sets the color of the formatted exception. **/
    public void setExceptionColor(String color) {
	exceptionColor = color;
	resetColor = PLAIN;
    }
    
    /** Sets the color of the formatted message. **/
    public void setMessageColor(String color) {
	messageColor = color;
	resetColor = PLAIN;
    }
    
    /** Sets the color of the formatted object. **/
    public void setObjectColor(String color) {
	objectColor = color;
	resetColor = PLAIN;
    }
    
    /** Sets the color of the formatted thread. **/
    public void setThreadColor(String color) {
	threadColor = color;
	resetColor = PLAIN;
    }
    
    /** Sets the color of the formatted classname. **/
    public void setClassnameColor(String color) {
	classnameColor = color;
	resetColor = PLAIN;
    }
    
    /** Sets the color of the formatted function. **/
    public void setFunctionColor(String color) {
	functionColor = color;
	resetColor = PLAIN;
    }
    
    /**
     * Sets color used to print the given event string in each log
     * line.  This means if the given event type is logged, the text
     * of the event type will be colored the given color.
     *
     * @param event to print in color
     * @param color that can be a concatenation of constant colors in
     * this class.
     **/
    public void setEventColor(String eventType, String color) {
	eventColors.put(eventType, color);
	resetColor = PLAIN;
    }

    /**
     * Sets color used to print an entire log line whose event type is
     * the given event string.
     *
     * @param event whose line to print in color
     * @param color that can be a concatenation of constant colors in
     * this class.
     **/
    public void setLineColor(String event, String color) {
	lineColors.put(event, color);
	resetColor = PLAIN;
    }
}

