//
// Copyright (C) 1999 Grace Software
//
// The contents of this file are subject to the Grace Software Public
// License Version 1.0 (the "GracePL"); you may not use this file
// except in compliance with the GracePL. You may obtain a copy of the
// GracePL at http://www.homestead.com/javalog/files/license.html
//
// Software distributed under the GracePL is distributed on an "AS IS"
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the GracePL for the specific language governing rights and
// limitations under the GracePL.
//
// The Original Code is Grace Software's JavaLog code, released March
// 21, 1999.
//
// The Initial Developer of the Original Code is Grace Software.
// Portions created by Grace Software are Copyright (C) 1999 Grace
// Software. All Rights Reserved.
//
// Contributor(s):
//
//
// Copyright (C) 1999 Grace Software
//
// The contents of this file are subject to the Grace Software Public
// License Version 1.0 (the "GracePL"); you may not use this file
// except in compliance with the GracePL. You may obtain a copy of the
// GracePL at http://www.homestead.com/javalog/files/license.html
//
// Software distributed under the GracePL is distributed on an "AS IS"
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the GracePL for the specific language governing rights and
// limitations under the GracePL.
//
// The Original Code is Grace Software's JavaLog code, released March
// 21, 1999.
//
// The Initial Developer of the Original Code is Grace Software.
// Portions created by Grace Software are Copyright (C) 1999 Grace
// Software. All Rights Reserved.
//
// Contributor(s):
//
package grace.log;

import java.io.File;
import java.io.PrintWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.BufferedReader;
import java.net.URL;
import java.rmi.RemoteException;


/**
 * This kind of PrintHandler writes all of its data into a file.  The
 * file can be specified in the constructor or as Properties.
 *
 * This Handler has the ability to write a limited length file instead
 * of a never ending file.  This is accomplished by monitoring the
 * size of the log file and when the file gets to a threshold size,
 * copy the portion of the file to a temporary file then rename the
 * temporary file to the original log file name.  This may cause havoc
 * with a log monitor program that reads the whole log and then waits
 * to read the asynchonous updates.  This type of program will have to
 * use of event sequence numbers to determine which events it has
 * already displayed.
 *
 * <P>Note that currently the events are always appended to the log
 * file - even between different runs of the logging program.  This
 * means that the user must manually remove the file if he wants the
 * log to start with a clean file.
 *
 * <p>The Properties interpreted by this class are:
 *
 * <pre>
 * log.handler.<i>name</i>.url = { file:///abs-path | file:rel-path }
 * log.handler.<i>name</i>.maxsize = 1000000 | 1000k | 1m
 * log.handler.<i>name</i>.file = filename <i>this is deprecated</i>
 * </pre>
 * 
 **/
public class FileHandler extends PrintHandler implements Handler {
    public static final String rcsid = "$Id: FileHandler.java,v 1.10 1999/10/11 00:10:00 cbladon Exp $";

    private File logFile;
    private long maxBytes = 0;

    private static URL contextURL;

    static {
	try {
	    contextURL = new URL("file://"
		+ System.getProperty("user.dir", ".") + "/");
	}
	catch (Exception e) {
	    e.printStackTrace();
	}
    }


    /**
     * This constructs a FileHandler that is initialized by
     * properties.
     **/
    public FileHandler(String prefix, String name) throws RemoteException
    {
	super(prefix, name);
	
	String propertyPrefix = prefix + name;

	//
	// First set up the file writer with the filename.  The
	// default name of the file is the property name + ".log".
	// This checks for a URL of the form file:///absolute-path or
	// file:relative-filename .
	//
	String filename = name + ".log";
	String urlString = System.getProperty(propertyPrefix + ".url",
	    "file:" + name + ".log");

	if (urlString.startsWith("file:")) {
	    try {
		URL fileURL = new URL(contextURL, urlString);
		filename = fileURL.getFile();
	    }
	    catch (Exception e) { e.printStackTrace(); }
	}
	
	establishWriter(filename);
	

	//
	// Now set up any limit on the file size.
	//
	String maxSizeString = System.getProperty(propertyPrefix + ".maxsize");

	if (maxSizeString != null) {
	    maxSizeString.trim();
	    maxSizeString = maxSizeString.toLowerCase();

	    int multiplier = 1;
	    if (maxSizeString.endsWith("k")) multiplier = 1000;
	    else if (maxSizeString.endsWith("m")) multiplier = 1000000;

	    try { maxBytes = Integer.parseInt(maxSizeString) * multiplier; }
	    catch (Exception e) { maxBytes = 0; }
	}
    }

    
    /**
     * This constructs a FileHandler that prints to the specified
     * file.
     *
     * @param filename to which to print log messages
     **/
    public FileHandler(String filename) throws RemoteException
    {
	super(System.out);    // temporary point to stdout until file is setup
	establishWriter(filename);
    }


    /**
     * This constructs a FileHandler that prints to the given filename
     * and limits the size of this file during logging to the given
     * maxKilobytes.  It does this by monitoring the size of the file,
     * copying a portion the file to a temporary file if it exceeds
     * maxKilobytes, and copying back to the log file.
     **/
    public FileHandler(String filename, int maxKilobytes)
	throws RemoteException
    {
	this(filename);
	maxBytes = maxKilobytes * 1024;
    }


    /**
     * This sets up the stream depending on what is given for
     * filename.  Filename can be a absoluate path, a relative path,
     * or '-' for standard out.  If there is any problem with the
     * opening the file, standard out is used.
     **/
    protected void establishWriter(String filename) {
	if (filename.endsWith("-")) {
	    logFile = null;
	    setOut(System.out);
	}
	else {
	    try {
		logFile = new File(filename);
		setOut(new PrintWriter(
		    new FileWriter(logFile.getAbsolutePath(), true), false));
	    }
	    catch (Exception e) {
		logFile = null;
		setOut(System.out);
	    }
	}
    }



    /**
     * The handle(Event) call is caught here so that the file size can be
     * trimmed after the event is written.
     **/
    public synchronized void handle(Event event) {
	super.handle(event);
	
	if (maxBytes > 0 && logFile != null && logFile.length() > maxBytes) {
	    cutFileInHalf();
	}
    }

    
    protected synchronized void cutFileInHalf() {
	if (logFile != null) {
	    try {
		getOut().close();

		BufferedReader reader =
		    new BufferedReader(new FileReader(logFile));

		//
		// Start near the end of the file and read until the
		// next newline.  Near means maxBytes / 2 from the
		// end.
		//
		reader.skip(logFile.length() - maxBytes / 2);
		while (reader.read() != '\n') {}

		//
		// Now copy the remaining entries to a temp file.
		//
		File tempFile = new File(".temp.log."
		    + new java.util.Date().getTime());
		FileWriter tempWriter = new FileWriter(tempFile);

		String line;
		while ((line = reader.readLine()) != null) {
		    tempWriter.write(line);
		    tempWriter.write('\n');
		}

		tempWriter.close();
		tempFile.renameTo(logFile);
	    }
	    catch (Exception e) {
	    }
	    
	    establishWriter(logFile.getName());
	}
    }


    public static void main(String[] args) {
	try {
	    File log = new File("test.log");
	    Log.getInstance().addHandler(new FileHandler(log.getName(), 1));

	    for (int i = 0; i < 200; ++i) {
		Log.log("test",
		    "test message " + i + " to force file over 1k");
		
		if (log.length() > 1024) {
		    System.out.println("warning: file exceeded 1K.");
		}
	    }
	}
	catch (Exception e) { e.printStackTrace(); }
    }
}
