//
// Copyright (C) 1999 Grace Software
//
// The contents of this file are subject to the Grace Software Public
// License Version 1.0 (the "GracePL"); you may not use this file
// except in compliance with the GracePL. You may obtain a copy of the
// GracePL at http://www.homestead.com/javalog/files/license.html
//
// Software distributed under the GracePL is distributed on an "AS IS"
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the GracePL for the specific language governing rights and
// limitations under the GracePL.
//
// The Original Code is Grace Software's JavaLog code, released March
// 21, 1999.
//
// The Initial Developer of the Original Code is Grace Software.
// Portions created by Grace Software are Copyright (C) 1999 Grace
// Software. All Rights Reserved.
//
// Contributor(s):
//
//
// Copyright (C) 1999 Grace Software
//
// The contents of this file are subject to the Grace Software Public
// License Version 1.0 (the "GracePL"); you may not use this file
// except in compliance with the GracePL. You may obtain a copy of the
// GracePL at http://www.homestead.com/javalog/files/license.html
//
// Software distributed under the GracePL is distributed on an "AS IS"
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the GracePL for the specific language governing rights and
// limitations under the GracePL.
//
// The Original Code is Grace Software's JavaLog code, released March
// 21, 1999.
//
// The Initial Developer of the Original Code is Grace Software.
// Portions created by Grace Software are Copyright (C) 1999 Grace
// Software. All Rights Reserved.
//
// Contributor(s):
//
package grace.util;

import java.io.StringReader;
import java.util.Date;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.TimeZone;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.text.ParsePosition;

import gnu.regexp.RE;
import gnu.regexp.REMatch;
import gnu.regexp.REException;

import grace.util.Time;
import grace.log.Log;
import grace.io.PrintWriter;



/**
 * Performs data scanning, gathering, and conversion functions on
 * text.  This class provides functions to read various canned object
 * like Strings, Dates, and integers as well as user defined object
 * types.  Most of the functionality uses regular expressions to
 * locate and delimit the text.
 *
 * <p>Objects of this class are stateful in that they maintain a
 * current position.  As data is parsed, this current position is
 * moved forwards or backwards.  Operations that get data from the
 * source typically set the current position in the source at the end
 * of the data returned.
 *
 * <p>Synopsis:
 * <pre>
 *   Tokenizer tokenizer = new Tokenizer("some text\nmore text\nnumCards=52");
 *   int numCards = tokenizer.getPrefixedInt("numCards=(\d+)", "$1");
 * </pre>
 *
 * <p>Notes:
 *
 * <p>If this class is used for scraping screens, the screen text
 * should contain any newlines that a meaningful to the format.  This
 * makes the * the screen scraping code more maintainable. In other
 * words, newlines should not be stripped out.  This will keep the
 * parsing code independent from the length of the lines in the text.
 * Therefore, if the text contains newlines and if the line length is
 * changed someday, the parsing code should not need to be changed.
 **/

public class Tokenizer {
    private String source;
    private int position = 0;
    private int lastPosition = 0;

    
    public Tokenizer(String source) {
	this.source = source;
    }


    /**
     * Useful if the source text should contain but doesn't.  This
     * probably means that, at one time, the source text had newlines,
     * but the newlines have been stripped out.  By injecting
     * newlines, the source text is more maintainably parsed because
     * the newlines add contextual information that was obviously
     * meant for human visualization of the data.
     *
     * @param lineLength periodic position in source at which newlines
     * should be injected
     **/
    public void injectNewlines(int lineLength) {
	StringBuffer newSource = new StringBuffer(source.length() + 100);

	int begin = 0;
	while (begin < source.length()) {
	    int end = begin + lineLength;
	    if (end > source.length()) end = source.length();

	    newSource.append(source.substring(begin, end));
	    newSource.append('\n');

	    begin = end;
	}

	source = newSource.toString();
    }
    

    /**
     * The copy allows the caller to capture the current state of this
     * Tokenizer such that this Tokenizer can continue parsing but not
     * affect the copy.  For example, this is useful to capture
     * <i>anchors</i> in input or perform different or parallel
     * parsing operations.
     **/
    public Object clone() { return this.clone(); }


    /**
     * Moves the current position to the next character in the source
     * that is not white as determined by
     * java.lang.Character.isWhitespace().
     **/
    public void skipWhiteSpace() {
	while (Character.isWhitespace(source.charAt(position))) ++position;
    }


    /**
     * Utility function that returns the first match in the source of
     * the given regular expression and sets the current position to
     * the end of the first match.  
     *
     * @param expression to match, return, and position after
     * @return first string matching expression in source
     **/
    protected String getMatch(RE expression) {
	String result = null;
	REMatch match = expression.getMatch(source, position);
	if (match != null) {
	    result = match.toString();
	    position = match.getEndIndex();
	}

	return result;
    }


    /**
     * Utility function to find the first match of the given
     * expression in the source, substitute the found match into the
     * given susbstitution string and return the result.  The
     * substitution string may contain plain text as well as the
     * symbols $0-$9 as dictated by the GNU regular expression
     * library.  $0 refers the entire match and $1 through $9 refer to
     * the first through the nineth matched sub expression.
     *
     * <p>The current position is positioned after the last matched
     * character - not the end of the matched sub expression in the
     * substitution, if any exists.
     *
     * @param expression to match
     * @param substitutionString into which the matched expression in
     * source is substituted.
     * 
     * @return result of substituting the matched string found in
     * source into the given substitutionString.
     **/
    protected String getSubstitutedMatch(
	RE expression,
	String substitutionString)
    {
	String result = null;
	REMatch match = expression.getMatch(source, position);
	if (match != null) {
	    result = match.substituteInto(substitutionString);
	    position = match.getEndIndex();
	}

	return result;
    }


    /**
     * Move the current position forward the given number of
     * characters.
     *
     * @param numCharacters to advance in the source
     **/
    public void advance(int numCharacters) {
	position += numCharacters;
    }

    /**
     * Advances the current position in the source to the start of the
     * match given expression.
     *
     * @exception REException if the given regularExpression is invalid
     * @param regularExpression to position at start of match
     **/
    public void advance(String regularExpression) throws REException {
	advance(new RE(regularExpression));
    }

    /**
     * Advances the current position in the source to the start of the
     * match of given expression.
     *
     * @param regularExpression to position at start of match
     **/
    public void advance(RE expression) {
	lastPosition = position;
	REMatch match = expression.getMatch(source, position);
	if (match != null) position = match.getStartIndex();
    }


    /**
     * Advances the current position in the source to the start of the
     * match of given numbered subexpression.  If the given expression
     * contains no subexpressions, the current position is placed at
     * the start of the matched expression.
     *
     * @param expression to match and position at sub expression
     * @param positionAtSubExpressionNumber number of sub expression
     * in expression to place current position.
     **/
    public void advance(String expression, int positionAtSubExpressionNumber)
	throws REException
    {
	advance(new RE(expression), positionAtSubExpressionNumber);
    }


    /**
     * Advances the current position in the source to the start of the
     * match of given numbered subexpression.  If the given expression
     * contains no subexpressions, the current position is placed at
     * the start of the matched expression.
     *
     * @param expression to match and position at sub expression
     * @param positionAtSubExpressionNumber number of sub expression
     * in expression to place current position.
     **/
    public void advance(RE expression, int positionAtSubExpressionNumber) {
	lastPosition = position;
	REMatch match = expression.getMatch(source, position);
	if (match != null) {
	    int subExpressionStart = match.getSubStartIndex(
		positionAtSubExpressionNumber);
	    position = (subExpressionStart == -1) ? 
		match.getStartIndex() : subExpressionStart;
	    Log.trace("advanced to", this);
	}
    }


    /**
     * Returns the first string in the input that matches the given
     * regular expression or null if there are none.
     *
     * @exception REException if regularExpression has syntax errors
     * @param regularExpression
     *
     * @return first string in input that matches the
     * regularExpression or null if there are no more tokens.
     **/
    public String find(String regularExpression) throws REException {
	return find(new RE(regularExpression));
    }

    /**
     * Returns the first string in the input that matches the given
     * regular expression or null if there are none.
     *
     * @param regularExpression
     *
     * @return first string in input that matches the
     * regularExpression or null if there are no more tokens.
     **/
    public String find(RE regularExpression) {
	lastPosition = position;
	return getMatch(regularExpression);
    }

    /**

     * Returns the result of substituting the result of the first
     * match of the given regularExpression into the given
     * substituteString.  Returns null if no match is made.  The
     * current position is set to next character after the end of
     * found string in source.
     *
     * <p>This function uses the gnu.regexp.REMatch.substituteInto()
     * function.  Therefore, the substitute string can contain plain
     * text or the special symbols $0-$9.  $0 represents the entire
     * matched string and $1-$9 represent the first thru the nineth
     * matched sub expression respectively.
     *
     * @see gnu.regexp.RE.getMatch(Object)
     * @see gnu.regexp.REMatch.substituteInto(String)
     *
     * @exception REException if regularExpression has syntax errors
     * @param regularExpression to match one time
     * @param substituteString into which match is substituted
     *
     * @return result of substitution of first match in input into the
     * substituteString or null if no match is made.
     **/
    public String findAndSubstitute(
	String regularExpression,
	String substituteString)
	throws REException
    {
	return findAndSubstitute(new RE(regularExpression), substituteString);
    }


    /**
     * Returns the result of substituting the result of the first
     * match of the given regularExpression into the given
     * substituteString.  Returns null if no match is made.  The
     * current position is set to next character after the end of
     * found string in source.
     *
     * <p>This function uses the gnu.regexp.REMatch.substituteInto()
     * function.  Therefore, the substitute string can contain plain
     * text or the special symbols $0-$9.  $0 represents the entire
     * matched string and $1-$9 represent the first thru the nineth
     * matched sub expression respectively.
     *
     * @see gnu.regexp.RE.getMatch(Object)
     * @see gnu.regexp.REMatch.substituteInto(String)
     *
     * @param regularExpression to match one time
     * @param substituteString into which match is substituted
     *
     * @return result of substitution of first match in input into the
     * substituteString or returns null if no match is made.
     **/
    public String findAndSubstitute(
	RE regularExpression,
	String substituteString)
    {
	lastPosition = position;
	return getSubstitutedMatch(regularExpression, substituteString);
    }


    /**
     * Move the current position backward the given number of
     * characters.
     *
     * @param numCharacters to advance in the source
     **/
    public void retreat(int numCharacters) {
	position -= numCharacters;
    }


    /**
     * Retreats the current position in the source to the start of the
     * given expression.
     *
     * @exception REException if the given regularExpression is invalid
     * @param regularExpression to position at start of match
     **/
    public void retreat(String regularExpression) throws REException {
	retreat(new RE(regularExpression));
    }

    /**
     * Retreats the current position in the source to the start of the
     * given expression.
     *
     * @param regularExpression to position at start of match
     **/
    public void retreat(RE expression) {
	lastPosition = position;
	REMatch match = expression.getMatch(source, position);
	if (match != null) position = match.getStartIndex();
    }

    
    /**
     * Returns the text that matches the given regularExpression that
     * is assumed to start at the current position in the input source
     * or returns null if at the end of the source.  The current
     * position is set to next character after the end of found string
     * in source.
     *
     * @return immediate match of regularExpression or null if end of source
     * @param regularExpression to match starting at current position
     **/
    public String get(String regularExpression) throws REException {
	return get(new RE(regularExpression), 0);
    }

    /**
     * Returns the text that matches the given regularExpression that
     * is assumed to start at the current position in the input source
     * or returns null if at the end of the source.  The current
     * position is set to next character after the end of found string
     * in source.
     *
     * @return immediate match of regularExpression or null if end of source
     * @param regularExpression to match starting at current position
     **/
    public String get(RE expression) {
	return get(expression, 0);
    }
    

    /**
     * Returns the next match not more than <code>maxOffset</code>
     * characters from the current position in the input stream or
     * null if there are no more tokens.  This function is good for
     * finding optional tokens that optionally appear in the source at
     * or near a fixed position.  The current position is set to next
     * character after the end of found string in source.
     *
     * @param regularExpression to match starting at current position
     * @param maxOffset maximum number of characters from current
     * position that match will succeed
     * @return matched string or null
     **/
    public String get(String regularExpression, int maxOffset)
	throws REException
    {
	return get(new RE(regularExpression, maxOffset));
    }

    /**
     * Returns the next match not more than <code>maxOffset</code>
     * characters from the current position in the input stream or
     * null if there are no more tokens.  This function is good for
     * finding optional tokens that optionally appear in the source at
     * or near a fixed position.  The current position is set to next
     * character after the end of found string in source.
     *
     * @param regularExpression to match starting at current position
     * @param maxOffset maximum number of characters from current
     * position that match will succeed
     * @return matched string or null
     **/
    public String get(RE expression, int maxOffset) {
	lastPosition = position;

	try {
	    REMatch match = expression.getMatch(source, position);
	    if (match.getStartIndex() <= position + maxOffset)
	    {
		position = match.getEndIndex();
		return match.toString();
	    }
	    else return "";
	}
	catch (Exception e) { return null; }
    }


    /**
     * Returns the next white space delimited token in the input
     * stream or null if there are no more tokens.  Here, white space
     * is a space, tab, or newline.  The current position is set to
     * next character after the end of token in source.
     *
     * @return next white space delimited token string or null if end
     * of source
     **/
    public String getToken() {
	lastPosition = position;

	try { return getMatch(new RE("\\S+")); }
	catch (REException e) { return null; }
    }
    

    /**
     * Returns the next white space delimited token not more than
     * <code>lastRelativeStartPosition</code> characters from the
     * current position in the input stream or null if there are no
     * more tokens.  This function is good for finding tokens that
     * optionally appear in the source at or near a fixed position.
     * The current position is set to next character after the end of
     * token in source.
     *
     * @param maxNumWhiteSpaceChars
     * @return next token or null if there are no more
     **/
    public String getToken(int maxNumWhiteSpaceChars) {
	lastPosition = position;

	try {
	    RE expression = new RE("\\S+");
	    REMatch match = expression.getMatch(source, position);
	    if (match.getStartIndex() <= position + maxNumWhiteSpaceChars)
	    {
		position = match.getEndIndex();
		return match.toString();
	    }
	    else return "";
	}
	catch (Exception e) { return null; }
    }


    /**
     * Returns a line of text (delimited by newline) from input
     * without the newline character in the result.  If the current
     * position is on a newline, the next full line is returned.
     * Otherwise, if the current position is in the middle of a line,
     * the remainder of the line is returned.  If there are no lines
     * left, null is returned.  The current position is set to start
     * of the next line in the source.
     *
     * @return next line of text without the newline character or null
     * if none
     **/
    public String getLine() {
	//
	// This is the lazy approach; it just returns null if any
	// exception is thrown.  This is meant to catch the end of the
	// input and anything else that goes wrong.
	//
	try {
	    lastPosition = position;
	    int end = position;

	    // Skip a newline if positioned on one
	    if (source.charAt(position) == '\n') ++position;

	    //
	    // Now find the end of the line
	    //
	    end = position;
	    while (end < source.length() && source.charAt(end) != '\n') ++end;

	    //
	    // check if we've hit the end of the source string
	    //
	    if (position == end) return null;
	    else {
		String result = source.substring(position, end);
		position = end + 1;	// passed the newline
		return result;
	    }
	}
	catch (Exception e) { return null; }
    }


    /**
     * Parses and returns the next token (skipping white space) as an
     * integer.  The current position is set to next character after
     * the end of int token in source.
     *
     * @exception NumberFormatException if the next token is not an integer
     * @return result of parsing the next token as an int
     **/
    public int getInt() throws NumberFormatException {
	return Integer.parseInt(getToken());
    }

    /**
     * Parses and returns the next token (skipping white space) as an
     * integer of the given maximum number of digits.  The current
     * position is set to next character after the end of int token in
     * source.
     *
     * @exception NumberFormatException if the next token is not an integer
     * @return result of parsing maxNumDigits of characters of the
     * next token as an int
     **/
    public int getInt(int maxNumDigits) throws NumberFormatException {
	try {
	    return Integer.parseInt(get("\\s*\\d{1," + maxNumDigits + "}"));
	}
	catch (REException e) { return Integer.MIN_VALUE; }
    }


    /**
     * This takes matches the given regular expression, sustitutes the
     * match into the given substitution string, parses the result as
     * an integer, and returns the result.  The current position is
     * set to next character after the end of int token in source.
     *
     * <p>This function uses the gnu.regexp.REMatch.substituteInto()
     * function.  Therefore, the substitute string can contain plain
     * text or the special symbols $0-$9.  $0 represents the entire
     * matched string and $1-$9 represent the first thru the nineth
     * matched sub expression respectively.
     *
     * @param expression to match in source
     * @param substitution into which matched expression is substituted
     * @exception NumberFormatException if the sustituted value is not an int
     * @return parsed integer result of subtituted match of given expression
     **/
    public int getInt(String expression, String substitution)
	throws NumberFormatException, REException
    {
	return getInt(new RE(expression), substitution);
	
    }

    /**
     * This matches the given regular expression, sustitutes the match
     * into the given substitution string, parses the result as an
     * integer, and returns the result as an integer.
     *
     * <p>This function uses the gnu.regexp.REMatch.substituteInto()
     * function.  Therefore, the substitute string can contain plain
     * text or the special symbols $0-$9.  $0 represents the entire
     * matched string and $1-$9 represent the first thru the nineth
     * matched sub expression respectively.
     *
     * @param expression to match in source
     * @param substitution into which matched expression is substituted
     * @exception NumberFormatException if the sustituted value is not an int
     * @return parsed integer result of subtituted match of given expression
     **/
    public int getInt(RE expression, String substitution)
	throws NumberFormatException
    {
	skipWhiteSpace();
	REMatch match = expression.getMatch(source, position);
	if (match != null && match.getStartIndex() == 0) {
	    String toConvert = match.substituteInto(substitution);
	    return Integer.parseInt(toConvert);
	}

	throw new NumberFormatException();
    }

    /**
     * Parses and returns the integer token <i>after</i> the matching
     * regular expression.  This is useful if the integer is prefixed
     * by some recogniseable string.  If an error occurs, the current
     * position is not changed.  The current position is set to next
     * character after the end of parsed int in source.
     *
     * @exception NumberFormatException if the next token is not an integer
     * @return parsed integer prefixed by given tagRegularExpression
     **/
    public int getPrefixedInt(String tagRegularExpression)
	throws NumberFormatException, REException
    {
	lastPosition = position;
	String token = getMatch(new RE(tagRegularExpression));
	return getInt();
    }

    /**
     * Parses and returns the next integer in the input (skipping non
     * decimal digits and white space).  The current position is set
     * to next character after the end of next int token in source.
     *
     * @exception ParseException if no integer in rest of input
     * @return next integer int source
     **/
    public int getNextInt() throws ParseException {
	lastPosition = position;
	String token = null;

	try { token = getMatch(new RE("\\d+")); }
	catch (REException e) {}
	
	if (token == null) {
	    throw new java.text.ParseException(
		"can't find an int", position);
	}
	else return Integer.parseInt(token);
    }


    /**
     * Parses and returns the next token as a Date parsed by a
     * SimpleDateFormat object created with the given simpleDateFormat
     * string.  The current position is set to next character after
     * the end of date in source.
     *
     * @exception ParseException if date failed to parse
     * @param simpleDateFormat to create a SimpleDateFormat to parse the date
     * @return next token parsed as a data according to given simpleDateFormat
     **/
    public Date getDate(String simpleDateFormat) throws ParseException {
	SimpleDateFormat format = new SimpleDateFormat(simpleDateFormat);
	format.setTimeZone(Time.gmt());
	return getDate(format);
    }
    
    /**
     * Parses and returns the next token as a Date parsed by the given
     * format.  The current position is set to next character after
     * the end of date in source.
     *
     * @exception ParseException if date failed to parse
     * @param format to parse the date
     * @return parsed Date from current position in source
     **/
    public Date getDate(DateFormat format) throws ParseException {
	lastPosition = position;

	skipWhiteSpace();
	
	ParsePosition parsePosition = new ParsePosition(position);
	Date result = format.parse(source, parsePosition);
	if (result != null) position = parsePosition.getIndex();
	else throw new ParseException("can't parse date", position);

	return result;
    }
    

    /**
     * Parses and returns the next token as a Date parsed by the given
     * time format but using the year, month, and date of the given
     * date.  Notice that because the date portion of the parsed time
     * is overwritten by the given date, only hours, minutes, and
     * seconds parsing sequences are useful in the given timeFormat.
     * The current position is set to next character after the end of
     * time in source.
     *
     * @exception ParseException if date failed to parse
     * @param format to parse the date
     * @param date to fill into result
     * @return parsed Date from current position in source
     **/
    public Date getTime(String timeFormat, Date date)
	throws ParseException
    {
	return getTime(new SimpleDateFormat(timeFormat), date);
    }
    

    /**
     * Parses and returns the next token as a Date parsed by the given
     * time format but using the year, month, and date of the given
     * date.  Notice that because the date portion of the parsed time
     * is overwritten by the given date, only hours, minutes, and
     * seconds parsing sequences are useful in the given timeFormat.
     * The current position is set to next character after the end of
     * time in source.
     *
     * @exception ParseException if date failed to parse
     * @param format to parse the date
     * @param date to fill into result
     * @return parsed Date from current position in source
     **/
    public Date getTime(DateFormat timeFormat, Date date)
	throws ParseException
    {
	lastPosition = position;

	skipWhiteSpace();
	
	ParsePosition parsePosition = new ParsePosition(position);
	Date time = timeFormat.parse(source, parsePosition);
	if (time != null) {
	    time = grace.util.Time.mergedDateAndTime(date, time);
	    position = parsePosition.getIndex();
	}
	else throw new ParseException("can't parse date", position);

	return time;
    }
    

    /**
     * Parses and returns the token, <i>after</i> the given regular
     * expression matches, as a Date parsed by a SimpleDateFormat
     * object created with the given simpleDateFormat string.  An
     * object representing the regular expression may or may not be
     * created.  The current position is set to next character after
     * the end of Date in source.
     *
     * @exception ParseException from SimpleDateFormat.parse()
     * @param regularExpression to match before parsing the date
     * @param simpleDateFormat to use to parse the date
     **/
    public Date getPrefixedDate(
	String regularExpression,
	String simpleDateFormat)
	throws ParseException, REException
    {
	return getPrefixedDate(new RE(regularExpression), simpleDateFormat);
    }
    

    /**
     * Parses and returns the token, <i>after</i> the given regular
     * expression matches, as a Date parsed by a SimpleDateFormat
     * object created with the given simpleDateFormat string.  The
     * current position is set to next character after the end of Date
     * in source.
     *
     * @exception ParseException from SimpleDateFormat.parse()
     * @param regularExpression to match before parsing the date
     * @param simpleDateFormat to use to parse the date
     **/
    public Date getPrefixedDate(
	RE regularExpression,
	String simpleDateFormat)
	throws ParseException
    {
	String tag = getMatch(regularExpression);
	return getDate(simpleDateFormat);
    }

    /**
     * Parses and returns the token, <i>after</i> the given regular
     * expression matches, as a Date parsed by the format object.  An
     * object representing the regular expression may or may not be
     * created.  The current position is set to next character after
     * the end of Date in source.
     *
     * @exception ParseException from SimpleDateFormat.parse()
     * @param regularExpression to match before parsing the date
     * @param format to use to parse the date
     **/
    public Date getPrefixedDate(
	String regularExpression,
	DateFormat format)
	throws ParseException, REException
    {
	return getPrefixedDate(new RE(regularExpression), format);
    }
    

    /**
     * Parses and returns the token, <i>after</i> the given regular
     * expression matches, as a Date parsed by the format object.  The
     * current position is set to next character after the end of Date
     * in source.
     *
     * @exception ParseException from SimpleDateFormat.parse()
     * @param regularExpression to match before parsing the date
     * @param format to use to parse the date
     **/
    public Date getPrefixedDate(
	RE regularExpression,
	DateFormat format)
	throws ParseException
    {
	String tag = getMatch(regularExpression);
	return getDate(format);
    }


    /**
     * Return the index of current position in the source.
     *
     * @return index of current position in source
     **/
    public int getPosition() { return position; }

    
    /**
     * Set the index of current position in the source.
     *
     * @param absolute index to set
     * @return index of current position in source
     **/
    public void setPosition(int absolute) {
	lastPosition = position;
	position = absolute;
    }

    /**
     * Returns the entire source.
     *
     * @return index of current position in source
     **/
    public String getSource() { return source; }


    /**
     * This moves the current position to the position before the
     * previous function was called.  This should work for all of the
     * public functions.  It simple moves the current position to the
     * position prior to the previous function.
     *
     * <p>This only works once.  In other words, currently only one
     * undo operation is kept.
     **/
    public void undoLast() { position = lastPosition; }


    /**
     * Indicates that the given regular expression will match the at
     * current position.  This is useful if one wants to check the
     * existance of optional text in the source at the current
     * position.  The current position is not moved.
     *
     * @param regularExpression to match at current position
     * @exception REException if regularExpression is invalid
     * @return given expression matches at current position
     **/
    public boolean isAt(String regularExpression)
	throws REException
    {
	return isAt(new RE(regularExpression));
    }
    
    /**
     * Indicates that the given regular expression will match the at
     * current position.  This is useful if one wants to check the
     * existance of optional text in the source at the current
     * position.  The current position is not moved.
     *
     * @param expression to match at current position
     * @return given expression matches at current position
     **/
    public boolean isAt(RE expression) {
	REMatch match = expression.getMatch(source, position);
	if (match != null && match.getStartIndex() == position) return true;
	else return false;
    }


    /**
     * Utility function to convert strings that have embedded non
     * printable characters such as newlines and tabs, and returns a
     * string that may be cleanly printed.  In other words, these
     * characters are expanded to their printable character appearance
     * like \n and \t.
     *
     * @param notPrintable string presumably containing tabs and newlines
     * @return printable version of notPrintable with pretty tabs and newlines
     **/ 
    protected String toPrintable(String notPrintable) {
	StringBuffer printable = new StringBuffer(notPrintable.length());
	for (int i = 0; i < notPrintable.length(); ++i) {
	    char c = notPrintable.charAt(i);
	    switch (c) {
		case '\n': printable.append("\\n"); break;
		case '\t': printable.append("\\t"); break;
		default: printable.append(c); break;
	    }
	}

	return printable.toString();
    }
    

    /**
     * Used by grace.io.PrintWriter to nicely print this.
     **/
    public void printTo(PrintWriter writer) {
	writer.print("context", toString());
    }


    public String toString() {
	int start = position - 10;
	int end = position + 10;

	return toPrintable(
	    source.substring((start > 0) ? start : 0, position)
	    + '[' + source.charAt(position) + ']'
	    + source.substring(
		position + 1,
		(end < source.length()) ? end : source.length()));
    }
    


    /**
     * Test program not quite completed.
     **/
    public static void main(String[] args) {
	try {
	    {
		//
		// getTime(time-format, date)
		//
		TimeZone.setDefault(Time.gmt());
		Tokenizer tok = new Tokenizer("12:34:56");
		Date base = new Date();
		Log.trace("base", base);
		Log.trace("parsed", tok.getTime("HH:mm:ss", base));
	    }
	}
	catch (Exception e) { Log.error(e); }
    }
}
